package ch.epfl.lpd.ids.client;

import java.util.Observable;
import java.util.Observer;
import java.util.Hashtable;
import java.net.InetSocketAddress;
import java.io.IOException;


import ch.epfl.lpd.ids.gui.ClientGUI;
import ch.epfl.lpd.ids.serialization.*;
import ch.epfl.lpd.ids.utils.ClientShutdownHook;
import ch.epfl.lpd.ids.network.ChatMulticast;
import ch.epfl.lpd.ids.network.IMulticast;
import ch.epfl.lpd.ids.utils.Constants;
import ch.epfl.lpd.ids.exceptions.*;
import ch.epfl.lpd.ids.membership.*;
import ch.epfl.lpd.ids.fd.*;
import ch.epfl.lpd.ids.rmulticast.*;

import java.net.SocketException;



/**
 * This class represents a chat client. A chat client is responsible
 * for creating its GUI as well as the failure detector, the multicast and 
 * reliable multicast layers. The multicast layer 
 * ({@link ch.epfl.lpd.ids.network.IMulticast}) is used to send connections and 
 * disconnections whereas the reliable multicast layer 
 * ({@link ch.epfl.lpd.ids.rmulticast.IReliableMulticast}) is used to send the 
 * messages. Furthermore, we initialize a membership manager 
 * ({@link ch.epfl.lpd.ids.membership.IMembershipManager}) that periodically 
 * sends heartbeats. 
 * @author Sebastien Baehni
 */
public class ChatClient extends Observable implements IChatClient, 
                                                      IFailureDetectorCallBack, 
                                                      IReliableMulticastCallBack,
                                                      IChatClientProperties {
    
    /**
     * The {@link ch.epfl.lpd.ids.gui.ClientGUI} that represents the
     * GUI of this chat client.
     */
    ClientGUI gui;
       
     /**
     * The username of this chat client.
     */
    String username = "";
    
    /**
     * The chat room the user is currently connected to.
     */
    String currentGroup = "";
    
    /**
     * The message dispatcher used to register objects interested in receiving 
     * messages.
     */
    IMessageDispatcher imd = null;
    /**
     * The membership manager used to send heartbeats messages.
     */
    IMembershipManager imm = null;
    /**
     * The failure detector that detects other crashed chat clients.
     */
    IFailureDetector ifd = null;
    /**
     * The multicast layer used to send connections and disconnections.
     */ 
    IMulticast im = null;
    /**
     * The reliable multicast layer to send reliably to all chat clients 
     * connected to the chat room of this client.
     */
    IReliableMulticast irm = null;
    
    /**
     * True if we are connected to a chat room.
     */
    boolean connected = false;
       
    /**
     * Creates a new instance of ChatClient. This constructor
     * basically creates the {@link ch.epfl.lpd.ids.gui.ClientGUI}, and
     * initialize the MVC.
     */
    public ChatClient() {        
        this.gui = new ClientGUI(this);
        this.addObserver(this.gui);              
    }
    
    /**
     * We decouple the creation of the GUI with the fact of showing it.
     */
    public void start() {
        this.gui.setVisible(true);        
    }
    
    /**
     * Used to retrieve the username of the client.
     * @return The username of the client.
     */
    public String getUsername() {
        return this.username;
    }
    
    /**
     * Used to retrieve the name of the chat room.
     * @return The name of the chat room.
     */
    public String getGroup() {
        return this.currentGroup;
    }
    
    /**
     * Method used to initialize the ChatClient. 
     * We initialize the multicast and reliable multicast layer, we create the 
     * message dispatcher, the membership manager as well as the failure detector. 
     * @param multicastAddress The multicast address to send all the messages to.
     * @param port The port of the multicast address.
     */
    protected void initialize(String multicastAddress, String port) {
        try {
            this.im = new ChatMulticast(ch.epfl.lpd.ids.utils.Constants.MAX_MSG_SIZE);
            this.im.init(new InetSocketAddress(multicastAddress, new Integer(port).intValue()));
            this.imd = new ChatDispatcher(this);
            this.imm = new ChatMembershipManager(this.im, this);
            this.ifd = new ChatFailureDetector();
            this.irm = new ChatReliableMulticast(this.im,this.ifd, this);            
        }
        catch (IllegalArgumentException iae) {
            iae.printStackTrace();
            ChatClient.printUsage();
            
        }
        catch (IOException ioe) {
            ioe.printStackTrace();
        }
        
    }
    
    /**
     * Method called by the {@link ch.epfl.lpd.ids.gui.ClientGUI} when
     * the user wants to connect to the server. This method sets correctly the
     * different callbacks between the membership manager, the failure 
     * detector, the multicast and reliable multicast layer. We finally start 
     * the membership and the failure detector and send a connection message.
     * @param username The username of the client.
     * @param groupName The name of the chat room.
     * @return True if the client was able to connect, false otherwise.
     */
    public boolean connect(String username, String groupName) {
        try {
            if (!connected) {
                this.username = username;
                this.currentGroup = groupName;                
                
                this.im.registerCallBack(this.imd);
                this.imd.registerDispatcherCallBack(this.ifd);
                this.imd.registerDispatcherCallBack(this.irm);
                this.ifd.registerFailureDetectorCallBack(this);                
                this.irm.registerReliableMulticastCallBack(this);
                
                this.ifd.start();
                this.imm.startMembership(Constants.ALIVE_TIMEOUT);
                
                connected =true;
                this.im.send(new ch.epfl.lpd.ids.serialization.Message(Constants.CONNECTION_MSG,this.username, this.currentGroup, "New Connection".getBytes()));
                return true;
            }
        }
        catch (MulticastInitializationException mie) {
            IMessage tmpMsg = new ch.epfl.lpd.ids.serialization.Message("Oups IMulticast not initialized.", "Error", "Error", mie.getMessage().getBytes());
            this.setChanged();            
            this.notifyObservers(tmpMsg);
            mie.printStackTrace();
        }
        catch (MessageTooBigException mtbe) {
            IMessage tmpMsg = new ch.epfl.lpd.ids.serialization.Message("Message too big.", "Error","Error", mtbe.getMessage().getBytes());
            this.setChanged();            
            this.notifyObservers(tmpMsg);
            mtbe.printStackTrace();
        }
        return false;
    }
    
    /**
     * Method called by the {@link ch.epfl.lpd.ids.gui.ClientGUI} when
     * the user wants to disconnect. We remove all the different callbacks 
     * previously registered with the connect method. We furthermore stop the
     * reliable multicast, the failure detector and the membership manager.
     * @return True if the client was successfully disconnected, false
     * otherwise.
     */
    public boolean disconnect(String username) {  
        try {
            if (connected) { 
                
                this.im.removeCallBack(this.imd);
                this.ifd.removeFailureDetectorCallBack(this);
                this.imd.removeDispatcherCallBack(this.ifd);
                this.imd.removeDispatcherCallBack(this.irm);
                this.irm.removeReliableMulticastCallBack(this);
                
                this.irm.stop();
                this.ifd.stop();
                this.imm.stopMembership();
                
                connected = false;
                
                this.im.send(new ch.epfl.lpd.ids.serialization.Message(Constants.DISCONNECTION_MSG, this.username, this.currentGroup, ("Disconnection from "+this.username).getBytes()));
                
                this.username = "";
                this.currentGroup = ""; 
                
                return true;
            }
        }
        catch (MulticastInitializationException mie) {
            IMessage tmpMsg = new ch.epfl.lpd.ids.serialization.Message("Oups IMulticast not initialized.","Error", "Error", mie.getMessage().getBytes());
            this.setChanged();            
            this.notifyObservers(tmpMsg);
            mie.printStackTrace();
        }
        catch (MessageTooBigException mtbe) {
            IMessage tmpMsg = new ch.epfl.lpd.ids.serialization.Message("Message too big.", "Error","Error", mtbe.getMessage().getBytes());
            this.setChanged();            
            this.notifyObservers(tmpMsg);
            mtbe.printStackTrace();
        }
        return false;
       }
    
    /**
     * Method called by the {@link ch.epfl.lpd.ids.gui.ClientGUI} when
     * the user wants to send a message to the other chat
     * clients. This method call the rsend method of the reliable multicast layer.
     * @param msg The message as a {@link ch.epfl.lpd.ids.serialization.IMessage}.
     */
    public void sendMessage(IMessage msg) {
        if (connected) {
            this.irm.rsend(msg);
        }
    }
    
    /**
     * Method called when (a) new user(s) (dis)connect(s) to this chat room.
     * This method is called by the failure detector.
     * @param newUsers The array containing the usernames of the clients 
     * connected to this chat room.
     */
    public void newUsers(String[] newUsers) {
        this.setChanged();
        this.notifyObservers(newUsers);
    }
    
    /**
     * Method called when a new message is received by the reliable multicast
     * layer. This method updates accordingly the GUI.
     * @param msg The {@link ch.epfl.lpd.ids.serialization.IMessage} received.
     */
    public void rdeliver(IMessage msg) {
        this.setChanged();            
        this.notifyObservers(msg);
    }
    
    /**
     * Method called by {@link ch.epfl.lpd.ids.gui.ClientGUI} when the
     * user close the chat GUI. We simply call the disconnect method (via
     * the client shutdown hook)
     */
    public void close() {        
        System.exit(0);
        this.im.close();
    }                       
        
    /**
     * Method that prints a help menu.
     */
    public static void printUsage() {
        System.out.println("Usage:");
        System.out.println("java ch.epfl.lpd.ids.ChatClient <multicastAddress> <port>");
        
    }
    
    /**
     * Main method used to start the application.  It creates an
     * instance of a ChatClient, initialize it and start it.
     * @param args The first argument represents the multicast address while
     * the second arguments represents the port number.
     */
    public static void main(String[] args) {
        if (args.length<2) {
            ChatClient.printUsage();
            System.exit(-1);
        }
        ChatClient cc = new ChatClient();
        cc.initialize(args[0],args[1]);
        cc.start();
        Runtime.getRuntime().addShutdownHook(new Thread(new ClientShutdownHook(cc, cc)));   
    }
    
}
