package ch.epfl.lpd.ids.fd;

import java.util.Hashtable;
import java.util.Timer;
import java.util.List;
import java.util.Collections;
import java.util.ArrayList;

import ch.epfl.lpd.ids.utils.Constants;
import ch.epfl.lpd.ids.serialization.IDispatcherCallBack;
import ch.epfl.lpd.ids.serialization.IMessage;
import ch.epfl.lpd.ids.utils.Constants;


/**
 * This class implements the specification of a failure detector. Basically,
 * after some time, if the failure has not received any message from a chat 
 * client, it decides that this client has crashed. The failure detector timeout
 * is specified in the file {@link ch.epfl.lpd.ids.utils.Constants}.
 * @author baehni
 */
public class ChatFailureDetector implements IFailureDetector, IFailureDetectorTaskCallBack {
    
    /**
     * The hashtable containing the chat client together with the last time 
     * (in millis) the failure detector has received a message from them.
     */
    Hashtable<String,Long> users = null;
    /**
     * A timer used to refresh the list of alive clients.
     */
    Timer timer = null;
    /**
     * The list of callbacks to notify when the list of alive clients changes.
     */
    List<IFailureDetectorCallBack> ifdcb=null;
    
    /**
     * The task used to remove crashed clients.
     */
    ChatFailureDetectorTask currentTask = null;
    
    
    /** 
     * Creates a new instance of ChatFailureDetector.
     * We simply initialize the list of chat clients and the list of callbacks.
     */
    public ChatFailureDetector() {
        this.users = new Hashtable<String, Long>();
        this.ifdcb = Collections.synchronizedList(new ArrayList<IFailureDetectorCallBack>());
    }
    
    /**
     * Starts the failure detector (i.e., schedule the task each FD_TIMEOUT ({@link ch.epfl.lpd.ids.utils.Constants}) 
     * which remove olds clients. 
     */
    public void start() {
        this.timer = new Timer();
        this.currentTask = new ChatFailureDetectorTask(this.users);
        this.currentTask.registerFailureDetectorTaskCallBack(this);
        this.timer.scheduleAtFixedRate(this.currentTask,1000,Constants.FD_TIMEOUT+1000);
    }
    
    /**
     * Stops the failure detector (i.e., stops the task previously scheduled with
     * a call to the start() method).
     */
    public void stop() {
        this.timer.cancel();
        this.users.clear();
        if (this.currentTask != null) this.currentTask.removeFailureDetectorTaskCallBack(this);
        this.timer.purge();
    }
    
    /**
     * Register a callback used to get notified when the list of alive users 
     * changes.
     * @param cb The callback that will get notified when the list of alive
     * users changes. 
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean registerFailureDetectorCallBack(IFailureDetectorCallBack cb) {
        return this.ifdcb.add(cb);
    }
    
    /**
     * Remove a callback.
     * @param cb The callback to remove. Once this callback is removed, it is 
     * no more notified if the alive user list changes. 
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean removeFailureDetectorCallBack(IFailureDetectorCallBack cb) {
        return this.ifdcb.remove(cb);
    }
    
    /**
     * Returns true if the user is alive.
     * @param username The name of the user we want to test to be alive.
     * @return True if the user is alive, false otherwise (or if no user with
     * this username is known).
     */
    public boolean alive(String username) {
        if (this.users.get(username) == null) return false;
        return true;
    }
    
    /**
     * Returns the list of alive users.
     * @return The list of alive users.
     */
    public String[] getAliveUsers() {
        return this.users.keySet().toArray(new String[0]);
    }
    
    /**
     * This method is called when a new message for the chat room the failure
     * detector is responsible of is received. If this message is a disconnection
     * message, the corresponding user is removed from the list of active clients.
     * Otherwise, the failure detector updates the user from which the message
     * was received with the current time. 
     * @param msg The {@link ch.epfl.lpd.ids.serialization.IMessage} received by a chat 
     * client. 
     */
    public void handle(IMessage msg) {
        boolean changed = false;  
        if (msg.getHeader().compareTo(Constants.DISCONNECTION_MSG)==0) {
            if (this.users.get(msg.getUsername())!=null) {
                this.users.remove(msg.getUsername());
                changed = true;
            }
        }
        else {
            if (this.users.get(msg.getUsername())==null) {
                changed = true;
            }
            this.users.put(msg.getUsername(),System.currentTimeMillis());
        }
        if (changed) this.changed();
    }
    
    /**
     * Called by the task if one or several chat clients were detected to be 
     * crashed. In turn this method notifies the registered callbacks. 
     */
    public void changed() {
        synchronized(this.ifdcb) {
            for (IFailureDetectorCallBack cb: this.ifdcb) {
                cb.newUsers(this.users.keySet().toArray(new String[0]));
            }
        }
    }
    
}
