
package ch.epfl.lpd.ids.fd;

import java.util.TimerTask;
import java.util.Hashtable;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.List;
import java.util.Collections;
import java.util.ArrayList;


import ch.epfl.lpd.ids.utils.Constants;

/**
 * This class tests periodically if the chat clients are crashed. Chat clients
 * are considered to be crashed if the failure has not received any message 
 * from them since FD_TIMEOUT ({@link ch.epfl.lpd.ids.utils.Constants}).
 * @author baehni
 */
public class ChatFailureDetectorTask extends TimerTask {
    
    /**
     * The hashtable containing the chat client together with the last time 
     * (in millis) the failure detector has received a message from them.
     */
    Hashtable<String,Long> users = null;
    
    /**
     * The list of callbacks to notify when the list of alive clients changes.
     */
    List<IFailureDetectorTaskCallBack> ifdtcb=null;
    
    
    /** Creates a new instance of ChatFailureDetectorTask. 
     * @param users The hashtable of users that are currently considered to be alive.
     */
    public ChatFailureDetectorTask(Hashtable<String,Long> users) {
        this.users = users;
        this.ifdtcb = Collections.synchronizedList(new ArrayList<IFailureDetectorTaskCallBack>());
    }
        
    /**
     * Registers a callback used to get notified when the list of alive users 
     * changes.
     * @param cb The callback that will get notified when the list of alive
     * users changes. 
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean registerFailureDetectorTaskCallBack(IFailureDetectorTaskCallBack cb) {
        return this.ifdtcb.add(cb);
    }
    
    /**
     * Removes a callback.
     * @param cb The callback to remove. Once this callback is removed, it is 
     * no more notified if the alive user list changes. 
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean removeFailureDetectorTaskCallBack(IFailureDetectorTaskCallBack cb) {
        return this.ifdtcb.remove(cb);
    }
    
    /**
     * Method called periodically by the timer of the failure detector.
     * We simply parse the list of active clients and checks how much time was
     * elapsed since the last message we receive from any of them. If this time
     * is greater than FD_TIMEOUT ({@link ch.epfl.lpd.ids.utils.Constants}), we 
     * remove the client from the list of active clients. If the list of active 
     * clients changes, we notify the different callbacks. 
     */
    public void run() {
        boolean changed = false;
        Enumeration<String> usersEnum = this.users.keys();
        while (usersEnum.hasMoreElements()) {
            String user = usersEnum.nextElement();
            long userTime = this.users.get(user);
            Calendar now = Calendar.getInstance();
            Calendar userCal = Calendar.getInstance();
            userCal.setTimeInMillis(userTime+Constants.FD_TIMEOUT);
            if (now.after(userCal)) {
                this.users.remove(user);
                changed = true;
            }
        }
        if (changed) {
            synchronized(this.ifdtcb) {
                for (IFailureDetectorTaskCallBack cb:this.ifdtcb) {
                    cb.changed();
                }
            }
        }
    }
    
    
}
