package ch.epfl.lpd.ids.fd;

import ch.epfl.lpd.ids.serialization.IDispatcherCallBack;

/**
 * The specification of the chat failure detector.
 * It must be possible to start and stop the failure detector. To retrieve the 
 * list of alive users as well as to know if a user is alive or not.
 * Finally we can register and remove callbacks to get notified when the list of
 * alive users changes.
 * @author Sebastien Baehni
 */
public interface IFailureDetector extends IDispatcherCallBack  {
    
    /**
     * Returns true if the user is alive.
     * @param username The name of the user we want to test to be alive.
     * @return True if the user is alive, false otherwise (or if no user with
     * this username is known).
     */
    public boolean alive(String username);
    
    /**
     * Returns the list of alive users.
     * @return The list of alive users.
     */
    public String[] getAliveUsers();
    
    /*
     * Starts the failure detector.
     */
    public void start();
    
    /*
     * Stops the failure detector.
     */
    public void stop();
    
    /**
     * Register a callback used to get notified when the list of alive users 
     * changes.
     * @param ifdcb The callback that will get notified when the list of alive
     * users changes. 
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean registerFailureDetectorCallBack(IFailureDetectorCallBack ifdcb);
    
    
    /**
     * Remove a callback.
     * @param ifdcb The callback to remove. Once this callback is removed, it is 
     * no more notified if the alive user list changes. 
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean removeFailureDetectorCallBack(IFailureDetectorCallBack ifdcb);
    
}
