package ch.epfl.lpd.ids.network;

import java.net.MulticastSocket;
import java.net.InetSocketAddress;
import java.net.DatagramPacket;

import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.net.SocketException;

import ch.epfl.lpd.ids.serialization.IMessage;
import ch.epfl.lpd.ids.exceptions.MulticastInitializationException;
import ch.epfl.lpd.ids.exceptions.MessageTooBigException;

/**
 * Implementation of the multicast layer. This layer is used to send AND receive
 * messages to/from a multicast socket. This specific socket is specified in the 
 * init() method. 
 * @author Sebastien Baehni
 */
public class ChatMulticast implements IMulticast, Runnable {
    
    /**
     * The underlying multicast socket to/from which message are sent/received.
     */
    MulticastSocket ms = null;
    /**
     * The inet address to send the message to.
     */
    InetSocketAddress isa = null;
    
    /**
     * The list of callback to notify when new messages are received.
     */
    List<IMulticastCallBack> callBacks = null;
    
    /**
     * The maximum length of a message. Usually this is set to {@link ch.epfl.lpd.ids.utils.Constants.MAX_MSG_SIZE}
     */
    int maxMsgLength = -1;
    
    /**
     * If we are ready to receive messages on the inet address specified in parameter of the init() method.
     */
    boolean started = false;
    
    /** 
     * Creates a new instance of ChatMulticast.
     * @param maxMsgLength The maximal length of a message.
     */
    public ChatMulticast(int maxMsgLength) {
        this.maxMsgLength = maxMsgLength;
    }
    
    /**
     * Initialize the multicast layer. Creates the list of callbacks, the multicast socket
     * and join the group of the multicast address. Finally starts the thread to receive
     * messages on this address.
     * @param isa The {@link java.net.InetSocketAddress} used to send messages to.
     * @throws IOException If an IOException occurs while initializing the layer.
     */
    public boolean init(InetSocketAddress isa) throws IOException {
        this.isa = isa;
        this.callBacks = Collections.synchronizedList(new ArrayList<IMulticastCallBack>());
        this.ms = new MulticastSocket(this.isa.getPort());
        this.ms.joinGroup(this.isa.getAddress());
        this.started = true;
        new Thread(this).start();
        return true;
    }
    
    /**
     * Register a callback that will be notified by this layer when a new message is received.
     * @param imcb The callback to register.
     * @throws MulticastInitializationException If the layer was not initialized before a 
     * call to this method is done.
     */
    public boolean registerCallBack(IMulticastCallBack imcb) throws MulticastInitializationException {
        if (this.callBacks == null) throw new MulticastInitializationException("ChatMulticast not initialized.");
        return this.callBacks.add(imcb);
    }
    
    /**
     * Remove a callback. After a call to this method, the callback in parameter does 
     * not receive anymore message from this layer.
     * @param imcb The callback to remove.
     * @throws MulticastInitializationException If the layer was not initialized before a 
     * call to this method is done.
     */
    public boolean removeCallBack(IMulticastCallBack imcb) throws MulticastInitializationException {
        if (this.callBacks == null) throw new MulticastInitializationException("ChatMulticast not initialized.");
        return this.callBacks.remove(imcb);
    }
    
    /**
     * Method used to send messages to the multicast address specified in the init()
     * method. The message that is sent is serialized with the native Java APIs. A
     * check is done to test if the size of the message is acceptable with respect to the maximal
     * size specified in the constructor.
     * @param msg The {@link ch.epfl.lpd.ids.serialization.IMessage} to send.
     * @throws MulticastInitializationException If the layer was not initialized before a 
     * call to this method is done.
     * @throws MessageTooBigException If the message we want to send is greater than 
     * MAX_MSG_SIZE ({@link ch.epfl.lpd.ids.utils.Constants})
     */
    public synchronized void send(IMessage msg) throws MulticastInitializationException,
                                          MessageTooBigException {
        try {
            if (this.ms == null) throw new MulticastInitializationException("ChatMulticast not initialized.");
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            ObjectOutputStream oos = new ObjectOutputStream(baos);
            oos.writeObject(msg);
            oos.flush();
            byte[] buffer = baos.toByteArray();
            if (buffer.length > maxMsgLength) throw new MessageTooBigException("The message you want to send is too big");
            this.ms.send(new DatagramPacket(buffer,buffer.length, this.isa));
            baos.close();
            oos.close();
            }
        catch (SocketException se) {
            se.printStackTrace();
        }
        catch (IOException ioe) {
            ioe.printStackTrace();
        }
    }
    
    /**
     * Notifies all the callbacks when a new message is received.
     * @param msg The {@link ch.epfl.lpd.ids.serialization.IMessage} received.
     */
    private void dispatch(IMessage msg) {
        synchronized(this.callBacks) {
            for (IMulticastCallBack imcb: this.callBacks) {
                imcb.handle(msg);
            }
        }
        
    }
    
    /**
     * Main method of the thread. This method is used to receive asynchronously 
     * the message on the multicast address specified in the init() method. This 
     * method called the dispatch method above. 
     */
    public void run() {
        try {
            while (started) {
                byte[] buffer = new byte[this.maxMsgLength];
                DatagramPacket dp = new DatagramPacket(buffer,buffer.length);
                this.ms.receive(dp);
                ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(dp.getData()));
                try {
                    IMessage msg = (IMessage)ois.readObject();                    
                    this.dispatch(msg);
                }
                catch (ClassCastException cce) {
                    cce.printStackTrace();
                    System.out.println("We have received an object which is not an IMessage, we drop it");
                }
                catch (ClassNotFoundException cnfe) {
                    cnfe.printStackTrace();
                    System.out.println("We have received an object which is not an IMessage, we drop it");
                }
                
            }
        }
        catch (IOException ioe) {
            System.out.println("Error while receiving a message. We close everything.");
            ioe.printStackTrace();
            this.close();
        }
    }

    /**
     * Closes the underlying multicast socket. We leave the group of the multicast
     * address and clear all the callbacks.
     */
    public void close() {
        try {
            this.ms.leaveGroup(this.isa.getAddress());
            }
        catch (IOException ioe) {
            ioe.printStackTrace();
        }
        this.isa = null;
        this.started = false;
        this.callBacks.clear();
        this.callBacks = null;
        this.ms.close();
        this.ms = null;
    }
    
   
}
