package ch.epfl.lpd.ids.network;

import java.net.MulticastSocket;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.io.IOException;

import ch.epfl.lpd.ids.serialization.IMessage;
import ch.epfl.lpd.ids.exceptions.MulticastInitializationException;
import ch.epfl.lpd.ids.exceptions.MessageTooBigException;

/**
 * Specification of the multicast layer. Before using the multicast layer it has
 * to be initialized through a call to the init() method. It is possible to register
 * callbacks in order to get notified when new messages are received. It is also possible
 * to send message (with the send()) method. Finally, the close() method closes the 
 * underlying multicast socket.
 * @author baehni
 */
public interface IMulticast {

    /**
     * Initialize the multicast layer.
     * @param isa The {@link java.net.InetSocketAddress} used to send messages to.
     * @throws IOException If an IOException occurs while initializing the layer.
     */
    public boolean init(InetSocketAddress isa) throws IOException;
    
    /**
     * Register a callback that will be notified by this layer when a new message is received.
     * @param imcb The callback to register.
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     * @throws MulticastInitializationException If the layer was not initialized before a 
     * call to this method is done.
     */
    public boolean registerCallBack(IMulticastCallBack imcb) throws MulticastInitializationException;
    
    /**
     * Remove a callback. After a call to this method, the callback in parameter does 
     * not receive anymore message from this layer.
     * @param imcb The callback to remove.
     * @throws MulticastInitializationException If the layer was not initialized before a 
     * call to this method is done.
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean removeCallBack(IMulticastCallBack imcb) throws MulticastInitializationException;
    
    /**
     * Method used to send messages to the multicast address specified in the init()
     * method. 
     * @param msg The {@link ch.epfl.lpd.ids.serialization.IMessage} to send.
     * @throws MulticastInitializationException If the layer was not initialized before a 
     * call to this method is done.
     * @throws MessageTooBigException If the message we want to send is greater than 
     * MAX_MSG_SIZE ({@link ch.epfl.lpd.ids.utils.Constants})
     */
    public void send(IMessage msg) throws MulticastInitializationException, MessageTooBigException;
    
    /**
     * Closes the underlying multicast socket.
     */
    public void close();
}
