package ch.epfl.lpd.ids.rmulticast;

import ch.epfl.lpd.ids.network.IMulticast;
import ch.epfl.lpd.ids.fd.IFailureDetector;
import ch.epfl.lpd.ids.serialization.IMessage;
import ch.epfl.lpd.ids.serialization.Message;
import ch.epfl.lpd.ids.utils.Constants;
import ch.epfl.lpd.ids.utils.Utils;
import ch.epfl.lpd.ids.exceptions.*;
import ch.epfl.lpd.ids.client.IChatClientProperties;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Timer;
import java.util.List;
import java.util.Collections;
import java.util.ArrayList;


/**
 * Implementation of the reliable multicast layer. We sends a message until it
 * has been acknowledged by all the correct clients of the chat room. When receiving a message, 
 * we acknowledge it and we rdeliver it only once.
 * @author Sebastien Baehni
 */
public class ChatReliableMulticast implements IReliableMulticast {
    
    /**
     * Timer used to re-send the messages not already acknowledged by all clients.
     */
    Timer timer = null;
    
    /**
     * The multicast layer.
     */
    IMulticast im=null;
    
    /**
     * The failure detector.
     */
    IFailureDetector ifd = null;
    
    /**
     * The hashtable containing the messages we have sent. The key represents the UID 
     * of the message.
     */
    Hashtable<Long,InternalMessage> messages = null;
    
    /**
     * To know if this layer has been started (i.e., re-sends messages).
     */
    boolean started = false;
    
    /**
     * The reference to the client properties.
     */
    IChatClientProperties iccp = null;
    
    /**
     * The list of callbacks to notify once when new messages are received.
     */
    List<IReliableMulticastCallBack> irmcb = null;
    
    /**
     * The list of the UIDs of the messages we have received. Used to prevent rdelivering twice the same
     * messages.
     */
    List<Long> UIDReceived = null;
    
    /** 
     * Creates a new instance of ChatReliableMulticast and initialize all the different
     * structures used in this class.
     * @param im The reference to the multicast layer.
     * @param ifd The reference to the failure detector.
     * @param iccp The reference to the properties of the client.
     */
    public ChatReliableMulticast(IMulticast im, IFailureDetector ifd, IChatClientProperties iccp) {
        this.im = im;
        this.ifd = ifd;
        this.iccp = iccp;
        this.messages = new Hashtable<Long,InternalMessage>();
        this.irmcb = Collections.synchronizedList(new ArrayList<IReliableMulticastCallBack>());
        this.UIDReceived = Collections.synchronizedList(new ArrayList<Long>());
    }
    
    /**
     * Sets the multicast layer used to (re-)sends the messages.
     * @param im The multicast layer. 
     */
    public void setMulticast(IMulticast im) {
        this.im = im;
    }
    
    /**
     * Sets the failure detector.
     * @param ifd The failure detector to deal with.
     */
    public void setFailureDetector(IFailureDetector ifd) {
        this.ifd = ifd;
    }
    
    /**
     * Sets several clients properties.
     * @param iccp The reference to the client properties.
     */ 
    public void setClientProperties(IChatClientProperties iccp) {
        this.iccp = iccp;
    }
    
    /**
     * Stops to send reliably messages. This stops the timer and clear all the
     * data structures.
     */
    public void stop() {
        if (this.timer != null) {
            this.timer.cancel();
            this.timer.purge();
        }
        this.messages.clear();
        this.UIDReceived.clear();
        this.started = false;
    }
    
    /** 
     * Sends reliably a message. The message is re-send until all the clients of 
     * the chat room has acknowledged it.
     * @param msg The {@link ch.epfl.lpd.ids.serialization.IMessage} to send.
     */
    public void rsend(IMessage msg) {
        if (!this.started) {
            this.started = true;
            this.timer = new Timer();
            this.timer.scheduleAtFixedRate(new ChatReliableMulticastTask(this.im,this.messages),1000,Constants.ACK_TIMEOUT);
        }
        this.messages.put(msg.getUID(),new InternalMessage(msg,this.ifd.getAliveUsers()));
        try {
            this.im.send(msg);
        }
        catch (MulticastInitializationException mie) {
            mie.printStackTrace();
        }
        catch (MessageTooBigException mtbe) {
            mtbe.printStackTrace();
        }
    }
    
    /**
     * This method is called by the message dispatcher when a new message is received.
     * If the message received is an acknowledgement, we remove the client that has acknowledged
     * from the list of the clients to send the message to. If the message received is a data message
     * we acknowledge it and rdeliver it once to the registered callbacks.
     * @param msg The {@link ch.epfl.lpd.ids.serialization.IMessage} received.
     */
    public void handle(IMessage msg) {
        if (msg.getHeader().compareTo(Constants.ACK_MSG)==0) {
            InternalMessage im = this.messages.get(msg.getUID());
            if (im != null) {
                im.removeUser(msg.getUsername());                
            }
        }
        else if (msg.getHeader().compareTo(Constants.DATA_MSG)==0) {
            synchronized(this.irmcb) {
                if (!this.UIDReceived.contains(msg.getUID())) {
                    for (IReliableMulticastCallBack cb: this.irmcb) {
                        cb.rdeliver(msg);
                    }
                    this.UIDReceived.add(msg.getUID());
                }                
            }

                        
            IMessage ackMsg = new Message(Constants.ACK_MSG,this.iccp.getUsername(), msg.getGroup(),"".getBytes());
            ackMsg.setUID(msg.getUID());
            try {
                // We wait a random time between 0 and RANDOM/1000 seconds in order
                // to prevent collision between the different acknowledgements.
               float random = Constants.RANDOM*Utils.nextFloat();
               Thread.currentThread().sleep((long)random);
               this.im.send(ackMsg);
            }
            catch (InterruptedException ie) {
                ie.printStackTrace();
            }
            catch (MulticastInitializationException mie) {
                mie.printStackTrace();
            }
            catch (MessageTooBigException mtbe) {
                mtbe.printStackTrace();
            }
            
        }
    }
    
    /**
     * Called by the failure detector when the list of active users changes.
     * We update accordingly the list of users to which the different messages must be sent to.
     * @param newUsers The list of active clients.
     */
    public void newUsers(String[] newUsers) {
        synchronized(this.messages) {
            Enumeration<InternalMessage> msgEnum = this.messages.elements();
            while (msgEnum.hasMoreElements()) {
                InternalMessage msg = msgEnum.nextElement();
                msg.updateUsers(newUsers);
            }
        }
    }
    
    /**
     * Register a callback that will be notified by this layer when a new message is received.
     * This callback is notified only ONCE even if the same is received multiple time.
     * @param cb The callback to register.
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean registerReliableMulticastCallBack(IReliableMulticastCallBack cb) {
        return this.irmcb.add(cb);
    }
    
    /**
     * Remove a callback. After a call to this method, the callback in parameter does 
     * not receive anymore message from this layer.
     * @param cb The callback to remove.
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean removeReliableMulticastCallBack(IReliableMulticastCallBack cb) {
        return this.irmcb.remove(cb);
    }
    
}
