package ch.epfl.lpd.ids.rmulticast;

import ch.epfl.lpd.ids.network.IMulticast;
import ch.epfl.lpd.ids.fd.IFailureDetector;
import ch.epfl.lpd.ids.fd.IFailureDetectorCallBack;
import ch.epfl.lpd.ids.serialization.IMessage;
import ch.epfl.lpd.ids.serialization.IDispatcherCallBack;
import ch.epfl.lpd.ids.client.IChatClientProperties;

/**
 * Specification of the reliable multicast layer. This layer uses the failure detector,
 * the multicast layer and several client properties. It is possible to register callbacks
 * in order to receive new message ONCE. Finally, this layer let a user call the rsend 
 * method to send reliably a message to all the users of the chat room according
 * to the failure detector. Note that if the client rsending the message crashes, no 
 * assumption is made. 
 * @author Sebastien Baehni
 */
public interface IReliableMulticast extends IDispatcherCallBack, IFailureDetectorCallBack  {

    /**
     * Stops to send reliably messages.
     */
    public void stop();
    
    /** 
     * Sends reliably a message.
     * @param msg The {@link ch.epfl.lpd.ids.serialization.IMessage} to send.
     */
    public void rsend(IMessage msg);
    
    /**
     * Sets the failure detector.
     * @param ifd The failure detector to deal with.
     */
    public void setFailureDetector(IFailureDetector ifd);
    
    /**
     * Sets the multicast layer used to (re-)sends the messages.
     * @param im The multicast layer. 
     */
    public void setMulticast(IMulticast im);
    
    /**
     * Sets several clients properties.
     * @param iccp The reference to the client properties.
     */ 
    public void setClientProperties(IChatClientProperties iccp);
    
    /**
     * Register a callback that will be notified by this layer when a new message is received.
     * This callback is notified only ONCE even if the same is received multiple time.
     * @param irmcb The callback to register.
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean registerReliableMulticastCallBack(IReliableMulticastCallBack irmcb);
    
    /**
     * Remove a callback. After a call to this method, the callback in parameter does 
     * not receive anymore message from this layer.
     * @param irmcb The callback to remove.
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean removeReliableMulticastCallBack(IReliableMulticastCallBack irmcb);
    
}
