package ch.epfl.lpd.ids.rmulticast;


import ch.epfl.lpd.ids.serialization.IMessage;


import java.util.List;
import java.util.Collections;
import java.util.ArrayList;

/**
 * Represents a message that needs to be retransmitted. This special structure
 * contains the message together with the list of clients that have not yet 
 * acknowledge it. Once every users have acknowledged the message, it does not need
 * to be retransmitted anymore.
 * @author Sebastien Baehni
 */
public class InternalMessage {
    
    /**
     * The reference to the message to resend.
     */
    IMessage msg = null;
    /**
     * The list of users that have not acknowledge this message.
     */
    List<String> users = null;
    
    /** 
     * Creates a new instance of InternalMessage.
     * @param msg The {@link ch.epfl.lpd.ids.serialization.IMessage} to resend.
     * @param users The array of users to resend the message to.
     */
    public InternalMessage(IMessage msg, String[] users) {
        this.msg = msg;
        this.users = Collections.synchronizedList(new ArrayList<String>());
        for (int i =0; i<users.length; i++) {
            this.users.add(users[i]);
        }
    }
    
    /**
     * Getter for the message.
     * @return The {@link ch.epfl.lpd.ids.serialization.IMessage} contained in this
     * structure.
     */
    public IMessage getMessage() {
        return this.msg;
    }
    
    /**
     * Updates the list of users to resend the message to.
     * @param users The new list of users to resend the message to.
     */
    public void updateUsers(String[] users) {
        this.users.clear();
        for (int i =0; i<users.length; i++) {
            this.users.add(users[i]);
        }
    }
    
    /**
     * Getter for the list of users to resend the message to.
     * @return The list of users to resend the message to.
     */
    public String[] getUsers() {
        return this.users.toArray(new String[0]);
    }
    
    /**
     * Removes a user from the list.
     * @param user A user that has either acknowledged the message or has crashed.
     */
    public void removeUser(String user) {
        this.users.remove(user);
    }
    
    /**
     * Test if this message needs to be resent.
     * @return True if the message need to be resent, false otherwise.
     */
    public boolean needToBeSent() {
        return this.users.size()!=0;
    }
   
}
