package ch.epfl.lpd.ids.serialization;

import java.util.List;
import java.util.Collections;
import java.util.ArrayList;

import ch.epfl.lpd.ids.client.IChatClientProperties;


/**
 * Implementation of a message dispatcher. This dispatchers notifies the registered
 * callbacks when a new message matching the users preferences (specified via the 
 * setClientProperties() method) is received.
 * @author Sebastien Baehni
 */
public class ChatDispatcher implements IMessageDispatcher {
    
    /**
     * The list of registered callbacks.
     */
    List<IDispatcherCallBack> idcb = null;
    
    /**
     * The reference to the client properties.
     */
    IChatClientProperties iccp = null;
    
    /** 
     * Creates a new instance of ChatDispatcher.
     * @param iccp The reference to the client properties.
     */
    public ChatDispatcher(IChatClientProperties iccp) {
        this.iccp = iccp;
        this.idcb = Collections.synchronizedList(new ArrayList<IDispatcherCallBack>());
    }
    
    /**
     * Sets several clients properties. These properties are used to check if the 
     * received messages need to be dispatched to the registered callbacks.
     * @param iccp The reference to the client properties.
     */
    public void setClientProperties(IChatClientProperties iccp) {
        this.iccp = iccp;
    }
    
    /**
     * Register a callback that will be notified by this layer when a new message is received.
     * This callback is notified only if the message match the client preferences.
     * @param idcb The callback to register.
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean registerDispatcherCallBack(IDispatcherCallBack idcb) {
        return this.idcb.add(idcb);
    }
    
    /**
     * Remove a callback. After a call to this method, the callback in parameter does 
     * not receive anymore message from the dispatcher.
     * @param idcb The callback to remove.
     * @return True if the collection used to store the callbacks has changed 
     * due to the call of this method.
     */
    public boolean removeDispatcherCallBack(IDispatcherCallBack idcb) {
        return this.idcb.remove(idcb);
    }
    
    /**
     * Method called by the multicast layer. 
     * We check if the received msg matches the client preferences and if this is the case
     * we notify the registered callback. In this application this means notifying the
     * failure detector and the reliable multicast layer.
     * @param msg The {@link ch.epfl.lpd.ids.serialization.IMessage} received.
     */
    public void handle(IMessage msg) {
        if (msg.getGroup().compareTo(this.iccp.getGroup())==0) {
            synchronized(this.idcb) {
                for (IDispatcherCallBack cb: this.idcb) {
                    cb.handle(msg);
                }
            }
        }
    }
    
    
    
    
    
}
