package ch.epfl.lpd.ids.serialization;

import ch.epfl.lpd.ids.utils.Utils;


/**
 * Class implementing the {@link
 * ch.epfl.lpd.ids.serialization.IMessage} specification.
 * @author Sebastien Baehni
 */
public class Message implements IMessage {

    /**
     * The header of the message.
     */
    String header = "";
    
    /**
     * The chat room in which this message needs to be sent.
     */
    String group = "";
    
    /**
     * The name of the sender.
     */
    String username = "";

    /**
     * The data of the message.
     */
    byte[] data;
    
    /** 
     * The unique identifer of this message.
     */
    long UID = Long.MIN_VALUE;


    /**
     * Creates a new instance of Message. This instance sets the
     * header and the data of the message.
     * @param header The header of the message.
     * @param username The name of the user that sends the message.
     * @param group The name of the chat room to which the message is sent to.
     * @param data The data of the message.
     */
    public Message(String header, String username, String group, byte[] data) {
        this.header = header;
        this.username = username;
        this.group = group;
        this.data = new byte[data.length];
        System.arraycopy(data, 0, this.data, 0, this.data.length);
        // We set an unique identifier 
        UID = Utils.getUID();
    }
    

    /**
     * Method used to retrieved the header of the message.
     * @return The header of the message.
     */
    public String getHeader() {
        return this.header;
    }
    
    /**
     * Method used to set the header of the message.
     * @param header The header of the message.
     */
    public void setHeader(String header) {
        this.header = header;
    }
    
    /**
     * Method used to retrieve the data of the message.
     * @return The data of the message.
     */
    public byte[] getData() {
        return this.data;
    }
    
    /**
     * Method used to set the data of the message.
     * @param data An array of bytes representing the data of the message.
     */
    public void setData(byte[] data) {
        System.arraycopy(data, 0, this.data, 0, this.data.length);
    }
    
    /**
     * Setter for the name of the chat room.
     * @param group The name of the room to which the message is sent to.
     */
    public void setGroup(String group) {
        this.group = group;
    }
    
    
    /**
     * Getter for the name of the chat room.
     * @return The name of the room to which the message is sent to.
     */
    public String getGroup() {
        return this.group;
    }
    
    
    /**
     * Setter for the name of the user that sends this message.
     * @param username The name of the user that sends this message.
     */
    public void setUsername(String username) {
        this.username = username;
    }
    
    
    /**
     * Getter for the name of the user that sends this message.
     * @return The name of the user that sends this message.
     */
    public String getUsername() {
        return this.username;
    }
    
    
    /**
     * Getter for the UID of this message.
     * @return The UID of this message.
     */
    public long getUID() {
        return this.UID;
    }
    
    
    /**
     * Setter for the UID of this message. Usually this method should never be called. 
     * The UID is set in the constructor. But for ack messages this is very useful.
     * @param uid The UID of the message.
     */
    public void setUID(long uid) {
        this.UID = uid;
    }
            
         
    /**
     * Method used to print this message as a string representation.
     * @return The string representation of this message.
     */
    public String toString() {
        return this.header+"\n"+this.username+"\n"+this.group+"\n"+new String(this.data);
    }
    
    
    
}
